/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

import metridoc.aggregator.BodyAggregator
import metridoc.aggregator.InflightAggregationWrapper
import metridoc.dsl.GroovyRouteBuilder
import metridoc.dsl.JobBuilder
import metridoc.dsl.RouteConfig
import metridoc.dsl.Services
import metridoc.impl.iterator.ExcelIteratorCreator
import metridoc.impl.iterator.LineIteratorProvider
import metridoc.processor.ClosureProcessor
import metridoc.registry.MetridocSimpleRegistry
import metridoc.sql.SqlExtension
import metridoc.thread.IsolatedThreadGroup
import org.apache.camel.CamelContext
import org.apache.camel.Expression
import org.apache.camel.builder.ExpressionBuilder
import org.apache.camel.impl.DefaultCamelContext
import org.apache.camel.impl.DefaultCamelContextNameStrategy
import org.apache.camel.model.ProcessorDefinition
import org.apache.camel.spi.BrowsableEndpoint
import org.apache.camel.spi.Registry
import metridoc.dsl.JobRunner
import metridoc.utils.ObjectUtils
import org.apache.camel.ProducerTemplate
import org.apache.camel.util.ServiceHelper
import metridoc.utils.SchemaUtils
import org.springframework.jdbc.datasource.embedded.EmbeddedDatabaseType
import org.springframework.jdbc.datasource.embedded.EmbeddedDatabaseBuilder
import metridoc.utils.PropertyUtils

ProcessorDefinition.metaClass.process = {filter ->
    if (filter instanceof Closure) {
        filter = new ClosureProcessor(filter)
    }
    delegate.process(filter);
}

ProcessorDefinition.metaClass.aggregateBody = {
    def expression = ExpressionBuilder.constantExpression(true)
    delegate.aggregate(expression, new InflightAggregationWrapper(new BodyAggregator())).
            completionSize(500).completionTimeout(500);
}

ProcessorDefinition.metaClass.aggregateBody = {int size, long timeout = 500 ->
    def expression = ExpressionBuilder.constantExpression(true)
    delegate.aggregate(expression, new InflightAggregationWrapper(new BodyAggregator())).
            completionSize(size).completionTimeout(timeout);
}

ProcessorDefinition.metaClass.splitByLine = {
    Expression bean = ExpressionBuilder.beanExpression(LineIteratorProvider.class, "create");
    delegate.split(bean).streaming()
}

ProcessorDefinition.metaClass.splitByXlsRecord = {
    Expression bean = ExpressionBuilder.beanExpression(ExcelIteratorCreator.class, "create");
    delegate.split(bean).streaming()
}

ProcessorDefinition.metaClass.splitByLine = {int threads, int maxPoolSize = 20, int maxQueueSize = 20 ->
    Expression bean = ExpressionBuilder.beanExpression(LineIteratorProvider.class, "create");
    delegate.split(bean).streaming().threads(threads, maxPoolSize).maxQueueSize(maxQueueSize)
}

JobRunner.metaClass.runRoute = {LinkedHashMap args = null, Closure closure ->
    def routeConfig = new RouteConfig()
    if (args != null) {
        routeConfig = new RouteConfig(args)
    }

    def routeBuilder = new GroovyRouteBuilder(route: closure,
            jobRunner: delegate, usePolling: routeConfig.usePolling)

    CamelContext context = manager.services[MetridocSimpleRegistry.CAMEL_CONTEXT]
    if (!context.status.started) {
        context.start()
    }

    context.addRoutes(routeBuilder)

    def routes = routeBuilder.routeCollection.routes
    def inflight = context.inflightRepository
    def boolean activityDetected = true
    def threadName = "${JobBuilder.METRIDOC_JOBS}.${jobName}.route"

    IsolatedThreadGroup.run(name: threadName as String, timeout: routeConfig.timeout) {

        while (activityDetected) {

            routes.each {route ->
                def fromDefinitions = route.inputs
                fromDefinitions.each {from ->
                    def uri = from.uri
                    manager.services[MetridocSimpleRegistry.URIS_TO_CHECK].add(uri)
                    def endpoint = context.getEndpoint(uri)
                    def size = inflight.size(endpoint)

                    if (size == 0) {
                        activityDetected = false
                    }

                    if (endpoint instanceof BrowsableEndpoint && !activityDetected) {
                        size = endpoint.exchanges.size()
                        if (size == 0) {
                            activityDetected = false
                        }
                    }
                }
            }

            if (activityDetected) {
                Thread.sleep(routeConfig.completionCheckPeriod)
            }
        }
    }

    if (routeBuilder.unCaughtException) {
        println "exception found"
        throw routeBuilder.unCaughtException
    }
}

//since we can't make another runRoute binding, we need to directly extend JobRunner
JobRunner.metaClass.runRoute = {String route, body = ObjectUtils.NULL, headers = [:] ->
    def camelContext = manager.services[MetridocSimpleRegistry.CAMEL_CONTEXT]
    ProducerTemplate template = camelContext.createProducerTemplate()
    def answer = template.requestBodyAndHeaders(route, body, headers);
    ServiceHelper.stopService(template)
    return answer
}

configureJobBuilder {LinkedHashMap args = null, JobBuilder jobBuilder ->

    def Services services = jobBuilder.services
    def Registry registry
    if (!services.containsKey(MetridocSimpleRegistry.CAMEL_REGISTRY)) {
        registry = new MetridocSimpleRegistry(services)
        services[MetridocSimpleRegistry.CAMEL_REGISTRY] = registry
    } else {
        registry = services[MetridocSimpleRegistry.CAMEL_REGISTRY]
    }

    if (!services.containsKey(MetridocSimpleRegistry.CAMEL_CONTEXT)) {
        def camelContext = new DefaultCamelContext(registry)
        camelContext.nameStrategy = new DefaultCamelContextNameStrategy("metridocCamel")
        services[MetridocSimpleRegistry.CAMEL_CONTEXT] = camelContext
    }

    def stopCamel = services[MetridocSimpleRegistry.STOP_CAMEL]
    if (stopCamel && stopCamel instanceof Boolean) {
        services[MetridocSimpleRegistry.STOP_CAMEL] = stopCamel
    } else {
        services[MetridocSimpleRegistry.STOP_CAMEL] = true
    }

    services[MetridocSimpleRegistry.URIS_TO_CHECK] = [] as Set
}

configureJobBuilder {LinkedHashMap args = null, JobBuilder jobBuilder ->
    if (args && args.jobs) {
        if (args.jobs instanceof Class) {
            jobBuilder.load(args.jobs)
        }

        if (args.jobs instanceof List) {
            args.jobs.each {job ->
                if (job instanceof Class) {
                    jobBuilder.load(job)
                }
            }
        }
    }
}

def isHook = {String prefix, String key, Object value ->
    key.startsWith(prefix) &&
            key.size() > 3 &&
            value instanceof Closure
}

def addHooks = {String prefix, LinkedHashMap args, JobBuilder jobBuilder ->

    if (args) {
        args.keySet().each {String key ->
            if (isHook(prefix, key, args[key])) {
                jobBuilder.services[key] = args[key]
            }
        }
    }
}

//add hooks
configureJobBuilder {LinkedHashMap args = null, JobBuilder jobBuilder ->
    addHooks("pre", args, jobBuilder)
    addHooks("post", args, jobBuilder)
}

onStop() {JobBuilder jobBuilder ->
    boolean stopCamel = jobBuilder.services[MetridocSimpleRegistry.STOP_CAMEL]
    Set urisToCheck = jobBuilder.services[MetridocSimpleRegistry.URIS_TO_CHECK]
    CamelContext camelContext = jobBuilder.services[MetridocSimpleRegistry.CAMEL_CONTEXT]

    if (stopCamel) {
        def activityDetected = !urisToCheck.empty

        while (activityDetected) {

            urisToCheck.each {uri ->
                def endpoint = camelContext.getEndpoint(uri)
                def inflight = camelContext.inflightRepository
                def size = inflight.size(endpoint)

                if (size == 0) {
                    activityDetected = false
                }

                if (endpoint instanceof BrowsableEndpoint && !activityDetected) {
                    size = endpoint.exchanges.size()
                    if (size == 0) {
                        activityDetected = false
                    }
                }
            }
        }

        if (camelContext != null) {
            camelContext.stop()
            camelContext = null
        }

    }
}

new SqlExtension().run()

JobRunner.metaClass.liquibase = {args ->
    def schemaUtils = new SchemaUtils(dataSource: args.dataSource)
    schemaUtils.runSchema(args.schema)
}

JobRunner.metaClass.dataSource = {args ->
    def utils = new SchemaUtils()
    utils.dbArgs = args
    return utils.dataSource
}

JobRunner.metaClass.embeddedDataSource = {
    return new EmbeddedDatabaseBuilder().setType(EmbeddedDatabaseType.H2).build()
}

JobRunner.metaClass.loadProperties = {String... propertyFiles ->
    def util = new PropertyUtils()
    util.loadProperties(manager.services, propertyFiles)
}

JobRunner.metaClass.getEndpoint = {String endpointUri ->
    return manager.services[MetridocSimpleRegistry.CAMEL_CONTEXT].getEndpoint(endpointUri)
}